﻿#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma version=2.20
#pragma IgorVersion=7
#pragma IndependentModule=UserProcLoader

#include <Resize Controls>

// This file should be installed in the Igor Procedures folder.

// If you're using Igor Pro 8 or later, the IgorExchange Projects
// Installer can be used to install this project and to provide
// notifications when new versions of this project are released.
// http: www.igorexchange.com/project/Updater
static strconstant ksShortTitle="UserProcLoader" // the project short title on IgorExchange
static constant kProjectID=21127 // the project node on IgorExchange

strconstant ksPackageName = UserProcLoader
strconstant ksPrefsFileName = UserProcLoaderPrefs.bin
constant kPrefsVersion = 100

strconstant ksExcludeFolders = ""
strconstant ksHide = "// Procedure Loader Hide"
// ksExcludeFolders is a list of folders to ignore - subfolders are also
// excluded. Paths start from Igor Pro User Files folder.
// e.g., ksExcludeFolders = "User Procedures:FolderToIgnore:;"
// ksHide defines a flag to hide a procedure file from the procedure
// loader. Developers can use this to hide helper files that are
// #included by the main package procedure file. The flag text should be
// copied and inserted, without change and on its own line, within the
// first 20 lines of the procedure.

// https://www.wavemetrics.com/user/tony

// Provides an easy way to load and unload procedure files from the User
// Procedures or Wavemetrics Procedures folders.

// Requires Igor Pro 8+ for full functionality.
// In Igor 7 it's possible to select procedures that cannot be unloaded.

// Settings
// Fast file search option is intolerant of semicolons in filenames.
// Mouseover listbox items to see path to file and compiler directives
// (#pragma & #include lines) displayed as listbox help.
// Open IM on Double Click allows display of an independent module
// procedure window on double-click or right-click on a 'loaded' listbox
// item.
// Show Hidden Procedures ignores the hide flag defined by ksHide.
// This option is provided for developers and should not usually be set.
// Don't Check Igor Version allows loading of procedures with 
// incompatible IgorVersion pragma. 

//#define debug

menu "File"
	"Load or Unload Procedures/1",/Q, MakeProcLoaderPanel()
end

function MakeProcLoaderPanel()
	
	KillWindow /Z LoaderPanel
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:procLoader
	DFREF dfr = root:Packages:procLoader
	
	// load package preferences
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)
	
	// prepare waves for two listboxes
	Make /O/N=0 dfr:w_loadedSelection /WAVE=w_loadedSelection, dfr:w_unloadedSelection /WAVE=w_unloadedSelection
	Make /O/N=0/T dfr:w_loadedProcs, dfr:w_unloadedProcs, dfr:w_loadedTitle, dfr:w_unloadedTitle, dfr:w_loadedHelp, dfr:w_unloadedHelp
	wave /SDFR=dfr/T w_loadedProcs, w_unloadedProcs, w_loadedTitle, w_unloadedTitle, w_loadedHelp, w_unloadedHelp
	w_loadedTitle = {"loaded"}; w_unloadedTitle = {"unloaded"}
				
	// create panel
	NewPanel /K=1/N=LoaderPanel/W=(prefs.win.left,prefs.win.top,prefs.win.right,prefs.win.bottom) as "Load or Unload Procedures"
	ModifyPanel /W=LoaderPanel noedit=1
	SetWindow LoaderPanel userdata(version) = num2str(GetThisVersion())
	
	// popup to select procedures folder on disk
	PopupMenu popupFolder, win=LoaderPanel, pos={10, 10}, value="User Procedures;Wavemetrics Procedures"
	PopupMenu popupFolder, win=LoaderPanel, Proc=loaderPopMenuProc
	
	// button to open Procedure Browser window (useful if you want to open procedure after loading)
	Button buttonBrowser, win=LoaderPanel, pos={420, 15}, size={15,14}, Picture=pb, labelBack=0, title=""
	Button buttonBrowser, win=LoaderPanel, Proc=loaderButtonProc, help={"Open Procedure Browser"}
	
	// button for settings
	Button buttonSettings, win=LoaderPanel, pos={450, 14}, size={15,15}, Picture=cog, labelBack=0, title=""
	Button buttonSettings, win=LoaderPanel, Proc=loaderButtonProc, help={"Change Settings"}
	
	// listbox for procedure files in selected folder that are not currently loaded
	ListBox listboxUnloaded, win=LoaderPanel, pos={10, 40},size={200, 270}, listwave=w_unloadedProcs
	ListBox listboxUnloaded, win=LoaderPanel, selwave=w_unloadedSelection, mode=9, focusring=0
	ListBox listboxUnloaded, win=LoaderPanel, fsize=12, Proc=loaderListBoxProc, titleWave=w_unloadedTitle
	#if (IgorVersion() >= 8)
	ListBox listboxUnloaded, win=LoaderPanel, helpWave=w_unloadedHelp
	#endif
	
	// button to be enabled when a valid selection is made from a listbox	
	Button buttonLoadUnload, win=LoaderPanel, pos={220,145},size={40,40},title="< >", Proc=loaderButtonProc, disable=2
	
	// listbox for procedure files in folder that ARE currently loaded
	ListBox listboxLoaded, win=LoaderPanel, mode=9, selwave=w_loadedSelection, pos={270, 40}, size={200, 270}
	ListBox listboxLoaded, win=LoaderPanel, editstyle=1, listwave=w_loadedProcs, fsize=12, Proc=loaderListBoxProc
	ListBox listboxLoaded, win=LoaderPanel, titleWave=w_loadedTitle, focusring=0
	#if (IgorVersion() >= 8)
	ListBox listboxLoaded, win=LoaderPanel, helpWave=w_loadedHelp
	#endif
	
	// popup menus to choose sort order for listboxes
	PopupMenu popupUnloadedSort, win=LoaderPanel, pos={10, 320}, value="Sort By Name;Sort by Path;Sort By Creation Date;Sort By Modification Date"
	PopupMenu popupUnloadedSort, win=LoaderPanel, Proc=loaderPopMenuProc, mode=prefs.sortUnloaded
	PopupMenu popupLoadedSort, win=LoaderPanel, pos={270, 320}, value="Sort By Name;Sort by Path;Sort By Creation Date;Sort By Modification Date"
	PopupMenu popupLoadedSort, win=LoaderPanel, Proc=loaderPopMenuProc, mode=prefs.sortLoaded
	
	// resizing userdata for controls
	PopupMenu popupFolder,win=LoaderPanel,userdata(ResizeControlsInfo) = A"!!,A.!!#;-!!#@^!!#<pz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	PopupMenu popupFolder,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#u:Duafnzzzzzzzzzzz"
	PopupMenu popupFolder,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	Button buttonBrowser,win=LoaderPanel,userdata(ResizeControlsInfo) = A"!!,I8!!#<(!!#<(!!#;mz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button buttonBrowser,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	Button buttonBrowser,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	Button buttonSettings,win=LoaderPanel,userdata(ResizeControlsInfo) = A"!!,IG!!#;m!!#<(!!#<(z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button buttonSettings,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	Button buttonSettings,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	ListBox listboxUnloaded,win=LoaderPanel,userdata(ResizeControlsInfo)= A"!!,A.!!#>.!!#AW!!#BAz!!#](Aon\"Qzzzzzzzzzzzzzz!!#`-A7TLfzz"
	ListBox listboxUnloaded,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	ListBox listboxUnloaded,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"	
	Button buttonLoadUnload,win=LoaderPanel,userdata(ResizeControlsInfo)=A"!!,Gl!!#@u!!#>.!!#>.z!!#`-A7TLfzzzzzzzzzzzzzz!!#r+D.OhkBk2=!z"
	Button buttonLoadUnload,win=LoaderPanel,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!\";f87cLJBQO4Szzzzzzzzzz"
	Button buttonLoadUnload,win=LoaderPanel,userdata(ResizeControlsInfo)+=A"zzz!!#r+D.Oh\\ASGdjF8u:@zzzzzzzzzzzz!!!"	
	ListBox listboxLoaded,win=LoaderPanel,userdata(ResizeControlsInfo)= A"!!,HB!!#>.!!#AW!!#BAz!!#`-A7TLfzzzzzzzzzzzzzz!!#o2B4uAeBE/#4z"
	ListBox listboxLoaded,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	ListBox listboxLoaded,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	PopupMenu popupUnloadedSort,win=LoaderPanel,userdata(ResizeControlsInfo) = A"!!,A.!!#BZ!!#@,!!#<pz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	PopupMenu popupUnloadedSort,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	PopupMenu popupUnloadedSort,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	PopupMenu popupLoadedSort,win=LoaderPanel,userdata(ResizeControlsInfo) = A"!!,HB!!#BZ!!#@<!!#<pz!!#`-A7TLfzzzzzzzzzzzzzz!!#`-A7TLfzz"
	PopupMenu popupLoadedSort,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	PopupMenu popupLoadedSort,win=LoaderPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"

	// resizing userdata for panel
	SetWindow LoaderPanel,userdata(ResizeControlsInfo) = A"!!*'\"z!!#CU!!#Bizzzzzzzzzzzzzzzzzzzzz"
	SetWindow LoaderPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzzzzzzzz"
	SetWindow LoaderPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzz!!!"
	
	// resizing panel hook
	SetWindow LoaderPanel hook(ResizeControls)=ResizeControls#ResizeControlsHook
	
	// package clean up hook
	SetWindow LoaderPanel hook(Cleanup)=UserProcLoader#HookCleanup
	
	// populate listbox waves
	UpdateListboxWaves()
	
	return 1
end

// panel hook function, clean up on quit
function HookCleanup(STRUCT WMWinHookStruct &s)	
	if (s.eventCode == 2)
		SaveWindowPosition(s.winName)
		KillWindow /Z LoaderPrefsPanel
		KillDataFolder /Z root:packages:procloader
	endif
	return 0
end

function LoaderPopMenuProc(STRUCT WMPopupAction &pa)
	if (pa.eventCode == 2) // mouseup
		// check for out-of-date panel
		if (cmpstr(GetUserData(pa.win, "", "version"), num2str(GetThisVersion())))
			MakeProcLoaderPanel()
			return 0
		endif
		UpdateListboxWaves()
	endif
	return 0
end

function LoaderButtonProc(STRUCT WMButtonAction &ba)
	if (ba.eventCode != 2) // mouseup
		return 0
	endif
	
	// check for out-of-date panel
	if (cmpstr(GetUserData(ba.win, "", "version"), num2str(GetThisVersion())))
		MakeProcLoaderPanel()
		return 0
	endif

	strswitch(ba.ctrlName)
		case "buttonBrowser":
			DoIgorMenu "Windows", "Procedure Browser"
			break
		case "buttonSettings":
			PrefsSetOptions()
			break	
		case "buttonLoadUnload":			
			// ba.userdata: 1 for load, 0 for unload
			IncludeListboxSelection(str2num(ba.userdata))
			break		
	endswitch
	return 0
end

function LoaderListBoxProc(STRUCT WMListboxAction &lba)	
	// eventCode 1 Mouse down; 2 Mouse up; 3 Double click; 
	// 4 Cell selection; Cell selection plus Shift key	
	if (lba.eventCode < 0  || lba.eventCode > 5)
		return 0
	endif	
	
	// check for out-of-date panel
	if (lba.eventCode==1 && cmpstr(GetUserData(lba.win, "", "version"), num2str(GetThisVersion())))
		Execute /Z/P "UserProcLoader#MakeProcLoaderPanel()"
		return 0
	endif
		
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)
	
	if (prefs.options&0x20 && (lba.eventCode==1 || lba.eventCode==2))
		DragAndDrop(lba)			
		return 0	
	endif
	
	if (lba.eventCode==4 && lba.eventMod&2^4) // right click in listbox cell
		string procWinStr = ""
		if (stringmatch(lba.ctrlName, "listboxUnloaded"))
			return 0 // allow context menu for loaded files
		endif
		procWinStr = lba.listWave[lba.row] + ".ipf"
		wave /T w_normalProcs = ListToTextWave(WinList("*",";","WIN:128,INDEPENDENTMODULE:0"), ";")
		string popupStr = "show " + procWinStr
				
		if (isInTextWave(procWinStr, w_normalProcs) == 0)
			// cell contains an independent module
			if (!(prefs.options & 8))
				// user will have to find another way to open :)
				popupStr = "(show " + procWinStr
			endif
		endif
		PopupContextualMenu popupStr
		if (V_flag)
			LoaderDisplayProc(lba.listWave[lba.row] + ".ipf")
		endif
		return 0
	endif
	
	if (lba.eventCode==3 && lba.row>-1 && lba.row<DimSize(lba.listWave, 0)) // double click on an item
		if (cmpstr(lba.ctrlName, "listboxUnloaded") == 0)
			IncludeListboxSelection(1) // load procs in unloaded listbox on double click
		else // try to display the proc window on double click in loaded listbox
			LoaderDisplayProc(lba.listWave[lba.row] + ".ipf")
		endif
	endif
		
	if (lba.eventCode==4 || lba.eventCode==5) // cell selection		
		// are we selecting procedures to load or to unload?	
		if (cmpstr(lba.ctrlName, "listboxUnloaded") == 0)
			wave otherSelWave = root:Packages:procLoader:w_loadedSelection
		else
			wave otherSelWave = root:Packages:procLoader:w_unloadedSelection
		endif
		
		// clear selection in other listbox
		if (numpnts(otherSelWave))
			otherSelWave = otherSelWave & ~9
		endif
		
		// don't allow disabled cells to be selected
		// (can be done by shift-click or using arrow keys)	
		lba.selWave = lba.selWave & 0x80 ? lba.selWave & ~9 : lba.selWave
	endif
	
	// enable buttons based on selection
	updateButtons()
	return 0
end

function DragAndDrop(STRUCT WMListboxAction &lba)
	
	if (!(lba.eventCode & 0x03)) // neither mouseup nor mousedown
		return 0
	endif
	
	variable f = 72 / PanelResolution(lba.win) // point/pixel
	int dragStarted = str2num(lba.userdata)
	
	int insert = (cmpstr(lba.ctrlName,"listboxUnloaded") == 0) // 1 for InsertInclude, 0 for DeleteInclude
	string otherListBox = SelectString(insert, "listboxUnloaded", "listboxLoaded")
	
	if (lba.eventCode==2 && dragStarted) // mouseup, drag completed
		// find whether mouse is within OTHER listbox
		if (isInControl(lba.mouseLoc, lba.win, otherListBox))
			IncludeListboxSelection(insert)
		endif
		ListBox $lba.ctrlName, win=$lba.win, userdata="0"
	endif

	if (lba.eventCode==1 && dragStarted==0) // mousedown, new drag
		if ( lba.row < 0 || lba.row >= (DimSize(lba.listWave, 0)) )
			return 0
		endif

		int i, j, startrow, endrow, dx, dy, buttondown, mouseH, mouseV
		ListBox $lba.ctrlName, win=$lba.win, userdata="1", mode=0 // this stops cell selection as mouse moves
		// userdata = "1" indicates dragging is active, cleared on mouseup
		// figure out visible rows
		ControlInfo /W=$lba.win $lba.ctrlName
		startrow = V_startRow
		endrow = min(numpnts(lba.selwave) - 1, startrow + ceil((V_height/f/V_rowHeight)-3)) // v_rowheight in pixels
		
		// create a titlebox for every visible selected item
		string DBname, strTitle
		j = 0
		dx = 3 // text offsets
		dy = 5
		for (i=startrow;i<endrow+1;i++)
			if (lba.selwave[i] & 9) // bit 0 or 3
				DBname = "DragBox" + num2str(j)
				sprintf strTitle, "\sa%+0.2d\x%+0.2d ", dx, dy
				TitleBox $DBname,win=$lba.win, title=strTitle+lba.listWave[i], pos={f*lba.ctrlRect.left,f*(lba.ctrlRect.top+(i-startrow+1)*V_rowHeight+1.5)}, labelBack=(0xA320,0xCDEB,0xFFCA)
				TitleBox $DBname,win=$lba.win, frame=0, fixedSize=1, size={f*(lba.ctrlRect.right-lba.ctrlRect.left),17}, fsize=12, disable=0
				j++
			endif
		endfor
		
		// save coordinates of other listbox
		ControlInfo /W=$lba.win $otherListBox
		STRUCT rect pixelRect
		pixelRect.left   = v_left / f // point -> pixel
		pixelRect.right  = v_right / f
		pixelRect.top    = v_top / f
		pixelRect.bottom = pixelRect.top + v_height / f

		// monitor mouse movement until mouseup
		do
			GetMouse /W=$lba.win
			buttondown = V_flag & 1
			dx = v_left - lba.mouseLoc.h
			dy = v_top - lba.mouseLoc.v
			// keep current mouse position updated as mouse moves
			lba.mouseLoc.h = v_left
			lba.mouseLoc.v = v_top
			
			// move titleboxes with mouse
			for (i=0; i<j; i++)
				TitleBox /Z $"DragBox"+num2str(i), win=$lba.win, pos+={dx,dy}
			endfor
			
			// activate other listbox when mouse is within control
			if (pointInRect(lba.mouseLoc, pixelRect))
				ListBox $otherListBox, win=$lba.win, focusRing=1
				ModifyControl $otherListBox activate
			else
				ModifyControl $lba.ctrlName activate
			endif
			
			DoUpdate /W=$lba.win
		while (buttondown)

		// clear titleboxes and return listbox to normal mode
		for (i=0; i<j; i++)
			KillControl /W=$lba.win $"DragBox"+num2str(i)
		endfor
		ListBox $otherListBox, win=$lba.win, focusRing=0
		ListBox $lba.ctrlName, win=$lba.win, mode=9 // listbox doesn't regain focus
		ModifyControl $lba.ctrlName activate
	endif	// end of drag
end

function isInControl(STRUCT point &mouse, string strWin, string strCtrl)
	ControlInfo /W=$strWin $strCtrl
	variable f = 72 / PanelResolution(strWin)
	variable hpoint = mouse.h * f // pixels -> points
	variable vpoint = mouse.v * f
	return ( hpoint>V_left && hpoint<(V_right) && vpoint>V_top && vpoint<(V_top+V_height) )
end

// point and rect structures must have same units
function pointInRect(STRUCT point &pnt, STRUCT rect &r)
	return ( pnt.h>r.left && pnt.h<r.right && pnt.v>r.top && pnt.v<r.bottom )
end

function LoaderDisplayProc(string strProc)
	
	DisplayProcedure /W=$strProc
	GetWindow /Z $strProc active
	
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)
	int AllowOpenIM = prefs.options & 8
	
	if (V_Value==0 && AllowOpenIM) // didn't succeed in displaying proc
		Execute "SetIgorOption IndependentModuleDev=?"
		NVAR v_sav = v_flag
		if (v_sav == 0)
			Execute "SetIgorOption IndependentModuleDev=1"
			DisplayProcedure /W=$strProc
			Execute "SetIgorOption IndependentModuleDev=0"
			GetWindow /Z $strProc active
		endif
	endif
	if (V_Value == 0)
		DoIgorMenu "Windows", "Procedure Browser"
	endif
end

function PrefsSetOptions()
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)

	string items = PopupItemString("Use Fast Search", prefs.options&0x01)
	items += PopupItemString("Show File Path on Mouseover", prefs.options&0x02, disable=IgorVersion()<8)
	items += PopupItemString("Scan for Compiler Directives", prefs.options&0x04, disable=IgorVersion()<8)
	items += PopupItemString("Open IM on Double Click", prefs.options&0x08)
	items += PopupItemString("Show Obsolete WM Procs", prefs.options&0x10)
	items += PopupItemString("Enable Drag and Drop", prefs.options&0x20)
	items += PopupItemString("Show Hidden Procedures", prefs.options&0x40)
	items += PopupItemString("Don't Check Igor Version", prefs.options&0x80)
	PopupContextualMenu items
	if (!V_Flag)
		return 0
	endif
	int bit = V_flag - 1
	prefs.options = prefs.options %^ 2^bit // toggle the bit
	
	SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
	if (bit > 0)
		if (bit==4 && WMProcsSelected()==0)
			return 0
		endif
		UpdateListboxWaves()
	endif
	return 1
end

function /S PopupItemString(string strItem, int checked, [int disable])
	disable = paramisdefault(disable) ? 0 : disable
	if (disable)
		return "\\M0:(:" + strItem + ";"
	elseif (checked)
		return "\\M0:!" + num2char(18) + ":" + strItem + ";"
	endif
	return strItem + ";"
end

function UpdateButtons()
	DFREF dfr = root:Packages:procLoader
	wave /SDFR=dfr w_loadedSelection, w_unloadedSelection
	
	// examine first bit of each point in selection waves
	int selections = 0
	if (numpnts(w_unloadedSelection))
		Duplicate /free/O w_unloadedSelection w_testSel
		w_testSel = w_testSel & 1
		selections = sum(w_testSel)
		if (selections)
			Button buttonLoadUnload, win=LoaderPanel, title=">", disable=0, userdata="1"
			return 0
		endif		
	endif

	selections = 0
	if (numpnts(w_loadedSelection))
		Duplicate /free/O w_loadedSelection w_testSel
		w_testSel = w_testSel & 1
		selections = sum(w_testSel)
		if (selections)
			Button buttonLoadUnload, win=LoaderPanel, title="<", disable=0, userdata="0"
			return 0
		endif
	endif
	Button buttonLoadUnload win=LoaderPanel, title="< >", disable=2, userdata=""
	return 0
end

// insert = 1 for InsertInclude, 0 for DeleteInclude
function IncludeListboxSelection(int insert)
	
	DFREF dfr = root:Packages:procLoader
	if (insert)
		wave w_sel = dfr:w_unloadedSelection
		wave /T w_list = dfr:w_unloadedProcs
	else
		wave w_sel = dfr:w_loadedSelection
		wave /T w_list = dfr:w_loadedProcs
	endif
	string cmd = SelectString(insert, "DELETEINCLUDE", "INSERTINCLUDE")
	ControlInfo /W=LoaderPanel popupFolder
	int isWMproc = v_value - 1
	string startquote = SelectString(isWMproc, "\"", "<")
	string endquote = SelectString(isWMproc, "\"", ">")
	string strExct = "", procList = ""
	int i
	
	for(i=0;i<numpnts(w_sel);i+=1)
		if (w_sel[i] & (1 + 8)) // bit 0 or 3
			sprintf strExct, "%s %s%s%s", cmd, startquote, w_list[i], endquote
			Execute/P/Q/Z strExct
			procList = AddListItem(w_list[i], procList)
		endif
	endfor
	sprintf cmd, "#SelectItems(%g,\"%s\")", insert, procList
	Execute/P/Q/Z "COMPILEPROCEDURES "
	Execute/P/Q/Z GetIndependentModuleName() + "#UpdateListboxWaves()" // queue an update for the listboxes
	Execute/P/Q/Z GetIndependentModuleName() + cmd // select the items that have just been loaded/unloaded
	return 1
end

// Search recursively for ipf files and return list of files as textwave.
// Slow method avoids IndexedFile(path, -1, string) and IndexedDir(path, -1, string)
// because semicolon is valid character in filenames!
function /wave GetProcsRecursive(int fast)
	
	int maxFolders = 1000, fullpath = 1
	int folderIndex, i
	string strFile = "", strFolder = ""
	Make /free/T/N=0 w_folders, w_allFiles
	
	w_folders = {SelectedFolder()} // search recursively starting in this folder
	
	string strExcludeFolders = ""
	for(i=0;i<ItemsInList(ksExcludeFolders);i+=1)
		strExcludeFolders += SpecialDirPath("Igor Pro User Files", 0, 0, 0) + RemoveEnding(StringFromList(i,ksExcludeFolders), ":") + ":;"
	endfor
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)
	if ((prefs.options&0x10) == 0 && WMProcsSelected())
		strExcludeFolders = SpecialDirPath("Igor Application", 0, 0, 0) + "WaveMetrics Procedures:Obsolete:;" + strExcludeFolders
	endif

	// w_folders will grow as subfolders are added
	for(folderIndex=0;folderIndex<numpnts(w_folders);folderIndex+=1)
		// check files in folderIndexth folder from w_folders
		NewPath /O/Q/Z tempPathFF, w_folders[folderIndex]
		
		if (fast) // don't worry about semicolons in filenames, or shortcuts ending in .ipf!
			// add list of ipf files in current folder to w_allFiles
			wave /T w = ListToTextWave(IndexedFile(tempPathFF, -1, ".ipf"),";")
			if (fullpath)
				w = w_folders[folderIndex] + w
			endif
			TextWaveConcatenate(w_allFiles, w)
			// add list of folders in current folder to w_folders
			wave /T w = ListToTextWave(IndexedDir(tempPathFF, -1, 1), ";")
			w += ":"
			TextWaveZapList(w, strExcludeFolders)
			TextWaveConcatenate(w_folders, w)

			#ifdef WINDOWS
			// get a list of shortcuts
			wave /T w = ListToTextWave(IndexedFile(tempPathFF, -1, ".lnk"),";")
			if (numpnts(w))
				w = ResolveAlias(w_folders[folderIndex] + w)
				Duplicate /free w w_f
				// keep shortcuts to files in w, shortcuts to folders in w_f
				if (fullpath == 0)
					w = ParseFilePath(0, w, ":", 1, 0)
				endif
				TextWaveStringMatch(w, "*.ipf")
				TextWaveConcatenate(w_allFiles, w)
				TextWaveStringMatch(w_f, "*:")
				TextWaveConcatenate(w_folders, w_f)
			endif
			#else // mac
			// get a list of file aliases
			wave /T w = ListToTextWave(IndexedFile(tempPathFF, -1, "alis"),";")
			if (numpnts(w))
				w = ResolveAlias(w_folders[folderIndex] + w)
				if (fullpath == 0)
					w = ParseFilePath(0, w, ":", 1, 0)
				endif
				TextWaveStringMatch(w, "*.ipf")
				TextWaveConcatenate(w_allFiles, w)
			endif
			// get a list of folder aliases
			wave /T w = ListToTextWave(IndexedFile(tempPathFF, -1, "fdrp"),";")
			if (numpnts(w))
				w = ResolveAlias(w_folders[folderIndex] + w)
				TextWaveZapBlanks(w)
				TextWaveConcatenate(w_folders, w)
			endif
			#endif
			continue // next folder
		endif
				
		// the slow way: loop through files in folder
		for (i=0;1;i+=1)
			strFile = IndexedFile(tempPathFF, i, "????")
			if (strlen(strFile) == 0)
				break
			endif
			GetFileFolderInfo /Q/Z w_folders[folderIndex] + strFile
			if (V_isAliasShortcut)
				strFile = ResolveAlias(w_folders[folderIndex] + strFile)
				if (stringmatch(strFile,"*:"))
					w_folders[numpnts(w_folders)] = {strFile}
				endif
				if (stringmatch(strFile, "*.ipf"))
					if (fullpath == 0)
						strFile = ParseFilePath(0, strFile, ":", 1, 0)
					endif
					w_allFiles[numpnts(w_allFiles)] = {strFile}
				endif
				continue // next file
			endif
			
			if (stringmatch(strFile, "*.ipf"))
				if (fullpath)
					strFile = w_folders[folderIndex] + strFile
				endif
				w_allFiles[numpnts(w_allFiles)] = {strFile}
				continue
			endif
		endfor // next file
		
		// add subfolders in current folder
		for (i=0;1;i+=1)
			strFolder = IndexedDir(tempPathFF, i, 1)
			if (strlen(strFolder) == 0)
				break
			endif
			if (FindListItem(strFolder + ":", strExcludeFolders) == -1)
				w_folders[numpnts(w_folders)] = {strFolder + ":"}
			endif
		endfor // next subfolder
		if (folderIndex == maxFolders)
			Print "Warning: GetProcsRecursive exceeding max iterations"
			break
		endif
	endfor // next folder
	
	KillPath /Z tempPathFF
	return w_allFiles
end

// appends w2 to w
function TextWaveConcatenate(wave /T w, wave /T w2)
	int oldPnts = numpnts(w), newPnts = numpnts(w2)
	if (newPnts)
		Redimension /N=(oldPnts+newPnts), w
		w[oldPnts,] = w2[p-oldPnts]
	endif
end

// keep only points of textwave matching strMatch
function TextWaveStringMatch(wave /T w, string strMatch)
	if (numpnts(w))
		w = SelectString(stringmatch(w,strMatch), "", w)
		TextWaveZapBlanks(w)
	endif
end

// Recursively resolve shortcuts to files/directories
// returns full path or an empty string if the file does not exist or the
// shortcut points to a non existing file/folder
// strPath is full path to file, folder or alias
function /S ResolveAlias(string strPath)
	GetFileFolderInfo /Q/Z RemoveEnding(strPath, ":")
	if (V_isAliasShortcut)
		return ResolveAlias(S_aliasPath)
	endif
	if (v_flag == 0)
		return strPath // path to a file or folder
	endif
	return ""
end

// returns path to folder selected in popup
// 1: User Procedures 2: Wavemetrics Procedures
function /S SelectedFolder()
	ControlInfo /W=LoaderPanel popupFolder
	if (V_flag < 1)
		return ""
	elseif (v_value == 1)
		return SpecialDirPath("Igor Pro User Files", 0, 0, 0) + "User Procedures:"
	endif
	return SpecialDirPath("Igor Application", 0, 0, 0) + "Wavemetrics Procedures:"
end

// returns truth that Wavemetrics Procedures is selected
function WMProcsSelected()
	ControlInfo /W=LoaderPanel popupFolder
	return (v_value == 2)
end

// strip trailing stuff in the output from winlist
function /S FilenameFromString(string str)
	if (numtype(strlen(str)) == 2)
		return ""
	endif
	SplitString /E=".*\.ipf" str // remove module names
	return RemoveEnding(s_value, ".ipf")
end

// extract quoted filename from include statement
function /S FilenameFromIncludeStatement(string str)
	if (numtype(strlen(str)) == 2)
		return ""
	endif
	ControlInfo /W=LoaderPanel popupFolder
	int isWMproc = v_value - 1
	string startquote = SelectString(isWMproc, "\"", "<")
	string endquote = SelectString(isWMproc, "\"", ">")
	int start, stop
	start = strsearch(str, startquote, 0)
	stop = strsearch(str, endquote, start+1)
	if (start==-1 || stop==-1)
		return ""
	endif
	return str[start+1,stop-1]
end

// truth that string str is a member of textwave w
function isInTextWave(string str, wave /T w)
	if (numtype(strlen(str)) == 2)
		return 0
	endif
	FindValue /TEXT=str/TXOP=4/Z w
	return (V_Value > -1)
end

// Remove empty points from textwave w (and optionally the corresponding
// points of waves in wave reference wave w_zapRefs). Does not check
// numbers of points in w_zapRefs waves.
function TextWaveZapBlanks(wave /T w, [wave /WAVE w_zapRefs])
	if (numpnts(w) == 0)
		return 0
	endif
	int i, j
	for (i=numpnts(w)-1;i>=0;i-=1)
		if (stringmatch(w[i], ""))
			DeletePoints i, 1, w
			if (WaveExists(w_zapRefs))
				for (j=0;j<numpnts(w_zapRefs);j+=1)
					DeletePoints i, 1, w_zapRefs[j]
				endfor
			endif
		endif
	endfor
	return numpnts(w)
end

// removes all instances of items in strList from text wave w.
// case insensitive.
function TextWaveZapList(wave /T w, string strList)
	if (numpnts(w)==0 || strlen(strList)==0)
		return 0
	endif
	int i
	for (i=numpnts(w)-1;i>=0;i-=1)
		if (FindListItem(w[i], strList, ";", 0, 0) > -1)
			DeletePoints i, 1, w
		endif
	endfor
	return numpnts(w)
end

function /S GetProcFilePath(string strWin)
	GetWindow /Z $strWin file
	return StringFromList(1, s_value) + StringFromList(0, s_value)
end

function UpdateListboxWaves()
	
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)
		
	DFREF dfr = root:Packages:procLoader
	wave /SDFR=dfr w_loadedSelection, w_unloadedSelection
	int numLoaded, numUnloaded, numIncluded
	int UnloadedSortOrder, LoadedSortOrder
	string strIgnore = ""
	string strIgnoreList = "~$*;" // list of filename patterns to ignore
	// IndexedFile ignores names starting with "._" but may find OS-hidden files
	int i, imax
	
	ControlInfo /W=LoaderPanel popupUnloadedSort
	UnloadedSortOrder = v_value - 1 // 0 by name, 1 by path, 2 by creation, 3 by mod
	ControlInfo /W=LoaderPanel popupLoadedSort
	LoadedSortOrder = v_value - 1
	
	// create a free text wave containing a list of procedure files (full filepath)
	// in the selected source folder
	wave /T w_unloadedPaths = GetProcsRecursive(prefs.options & 1)
	
	numUnloaded = numpnts(w_unloadedPaths)
	Make /free/N=(numUnloaded)/T w_unloadedProcs
	
	// reduce file path to filename without extension
	w_unloadedProcs = ParseFilePath(3, w_unloadedPaths, ":", 0, 0)
	
	// remove names that match patterns in the ignore list
	imax = ItemsInList(strIgnoreList)
	for(i=0;i<imax;i+=1)
		strIgnore = StringFromList(i, strIgnoreList)
		w_unloadedProcs = SelectString(stringmatch(w_unloadedProcs, strIgnore), w_unloadedProcs, "")
	endfor
	
	// create a free text wave containing a list of all procedure files in the experiment
	// use colon as list separator because semicolon is legal in filenames
	wave /T w_loadedProcs = ListToTextWave(WinList("*",":","WIN:128,INDEPENDENTMODULE:1"),":")
	numLoaded = numpnts(w_loadedProcs)
	
	Make /free/N=(numLoaded)/T w_loadedPaths
	w_loadedPaths = GetProcFilePath(w_loadedProcs)

	// clean up names
	if (numLoaded)
		w_loadedProcs = FilenameFromString(w_loadedProcs) // some points may be blank
	endif
	
	// remove procedures that aren't in selected source folder from loaded list
	if (numLoaded)
		w_loadedProcs = SelectString(isInTextWave(w_loadedProcs[p], w_unloadedProcs), "", w_loadedProcs[p])
		numLoaded = TextWaveZapBlanks(w_loadedProcs, w_zapRefs={w_loadedPaths})
	endif
	
	// remove already loaded procedures from unloaded list
	if (numUnloaded)
		w_unloadedProcs = SelectString(isInTextWave(w_unloadedProcs[p], w_loadedProcs), w_unloadedProcs[p], "")
	  	numUnloaded = TextWaveZapBlanks(w_unloadedProcs, w_zapRefs={w_unloadedPaths})
	endif
	
	// all of these waves must remain synched with path and procname waves if we reorder or delete points
	Make /free/N=(numLoaded)/T w_loadedHeader, w_loadedHelp
	Make /free/N=(numLoaded)/D w_loadedDate
	Make /free/N=(numUnloaded)/T w_unloadedHeader, w_unloadedHelp
	Make /free/N=(numUnloaded)/D w_unloadedDate
	
	if (numUnloaded>0)
		w_unloadedHeader = GetHeader(w_unloadedPaths, prefs.options)
		w_unloadedHelp = Header2Help(w_unloadedHeader, w_unloadedPaths, prefs.options)
		w_unloadedPaths = SelectString(cmpstr(w_unloadedHelp, "HIDE")==0, w_unloadedPaths, "")
		numUnloaded = TextWaveZapBlanks(w_unloadedPaths, w_zapRefs={w_unloadedProcs, w_unloadedHelp, w_unloadedHeader, w_unloadedDate})
	endif
	if (numLoaded>0)
		w_loadedHeader = GetHeader(w_loadedPaths, prefs.options)
		w_loadedHelp = Header2Help(w_loadedHeader, w_loadedPaths, prefs.options)
		w_LoadedPaths = SelectString(cmpstr(w_loadedHelp, "HIDE")==0, w_LoadedPaths, "")
		numLoaded = TextWaveZapBlanks(w_LoadedPaths, w_zapRefs={w_loadedProcs, w_loadedHelp, w_loadedHeader, w_loadedDate})
	endif
	
	// sort order 0 by name, 1 by path, 2 by creation, 3 by mod
	if (unloadedSortOrder>1 && numUnloaded>0)
		w_unloadedDate = GetFileDate(w_unloadedPaths, unloadedSortOrder-2)
	endif
	if (loadedSortOrder>1 && numLoaded>0)
		w_loadedDate = GetFileDate(w_loadedPaths, loadedSortOrder-2)
	endif
	
	// figure out which procedures can be unloaded.
	// using INCLUDE:2 with winlist gets all included files,
	// not just those that we can DELETEINCLUDE
	
	// load procedure file lines into text wave
	wave /T w_includedProcs = ListToTextWave(ProcedureText("",0,"procedure"), "\r")
	
	// find lines that start with #include
	Grep /E="(?i)^#include[\s]*[<\"]"/Z w_includedProcs as w_includedProcs
	numIncluded = TextWaveZapBlanks(w_includedProcs)
	
	// extract file names from between quotes
	if (numIncluded)
		w_includedProcs = FilenameFromIncludeStatement(w_includedProcs)
	endif
  	
  	if (LoadedSortOrder>1 && numpnts(w_loadedDate)==numLoaded) // sort by creation/mod date
  		Sort/R w_loadedDate, w_loadedProcs, w_loadedHelp, w_loadedHeader, w_loadedPaths, w_loadedDate
  	elseif (LoadedSortOrder == 1) // sort by path
  		Sort w_loadedPaths, w_loadedProcs, w_loadedHelp, w_loadedHeader, w_loadedPaths
	else // sort by name
		Sort w_loadedProcs, w_loadedProcs, w_loadedHelp, w_loadedHeader, w_loadedPaths
	endif
  	
  	if (UnloadedSortOrder>1 && numpnts(w_unloadedDate)==numUnloaded)
  		Sort/R w_unloadedDate, w_unloadedProcs, w_unloadedHelp, w_unloadedHeader, w_unloadedPaths, w_unloadedDate
  	elseif (UnloadedSortOrder == 1)
		Sort w_unloadedPaths, w_unloadedProcs, w_unloadedHelp, w_unloadedHeader, w_unloadedPaths
  	else
		Sort w_unloadedProcs, w_unloadedProcs, w_unloadedHelp, w_unloadedHeader, w_unloadedPaths
	endif
	
	// copy free waves into the package folder
	Duplicate /O/T w_unloadedProcs dfr:w_unloadedProcs
	Duplicate /O/T w_unloadedHelp dfr:w_unloadedHelp
		
	Duplicate /O/T w_loadedProcs dfr:w_loadedProcs
	Duplicate /O/T w_loadedHelp dfr:w_loadedHelp
		
	#ifdef debug
	Duplicate /O w_unloadedDate dfr:w_unloadedDate
	Duplicate /O w_unloadedPaths dfr:w_unloadedPaths
	Duplicate /O w_loadedDate dfr:w_loadedDate
	Duplicate /O w_loadedPaths dfr:w_loadedPaths
	#endif
	
	Redimension /N=(numLoaded) w_loadedSelection
	Redimension /N=(numUnloaded) w_unloadedSelection
	
	// disable cells containing procedures that cannot be unloaded using DELETEINCLUDE
	if (numLoaded)
		w_loadedSelection = (1 - isInTextWave(w_loadedProcs, w_includedProcs)) * 0x80
	endif
	// disable cells containing procedures that require a more recent version of Igor
	if (numUnloaded)
		if (prefs.options & 0x80)
			w_unloadedSelection = 0
		else			
			w_unloadedSelection = IgorTooOld(w_unloadedHeader) * 0x80
		endif
	endif
	
	// nothing is selected, so disable load/unload button
	Button buttonLoadUnload win=LoaderPanel, title="< >", disable=2, userdata=""

	return 1
end

function IgorTooOld(string strHeader)
	wave /T wHeader = ListToTextWave(strHeader + "\r", "\r")
	string strExp = "(?i)^#pragma[\s]*IgorVersion[\s]*="
	Grep /Q/E=strExp/LIST/Z wHeader
	variable version
	sscanf LowerStr(s_value), "#pragma igorversion = %g", version
	return v_flag ? version > IgorVersion() : 0
end

function /S GetHeader(string strFile, int options)
		
	if (strlen(strFile) == 0)
		return "no file"
	endif
	
	int maxLines = 20 // number of lines in each file to search for compiler directives
	int refNum, i
	string strHeader = "", strLine = ""
	
	Open /R/Z refnum as strFile
	if (refnum == 0)
		return strHeader + "cannot open"
	endif
	for (i=0;i<maxLines;i+=1)
		FReadLine refNum, strLine
		if (cmpstr(strLine[0,3], "#pra")==0 || cmpstr(strLine[0,3], "#inc")==0)
			strHeader += strLine
		elseif (strlen(ksHide) && cmpstr(strLine[0, strlen(ksHide)-1], ksHide) == 0)
			if (!(options & 0x40))
				Close refnum
				return "HIDE"
			endif
			strHeader += strLine
		endif
	endfor
	Close refnum
	strHeader = SelectString(strlen(strHeader)==0, strHeader, "no pragma")
	return RemoveEnding(strHeader, "\r")
end

// options bit 1: full path to file, bit 2: #pragma and #include lines
// ignore other bits
function /S Header2Help(string strHeader, string strPath, int options)
	if (cmpstr(strHeader, "HIDE") == 0)
		return "HIDE"
	endif
	string strHelp = ""
	if (options & 2)
		string UserFiles = SpecialDirPath("Igor Pro User Files", 0, 0, 0)
		string IgorApp = SpecialDirPath("Igor Application", 0, 0, 0)
		if (stringmatch(strPath, UserFiles + "*"))
			strHelp = strPath[strlen(UserFiles), Inf] + "\r"
		elseif (stringmatch(strPath, IgorApp + "*"))
			strHelp = strPath[strlen(IgorApp), Inf] + "\r"
		else
			strHelp = strPath + "\r"
		endif
	endif
	if (options & 4)
		strHelp += strHeader
	endif
	return strHelp
end

function GetFileDate(string strFile, int modOrCreation)
	// modOrCreation 0: creation, 1: mod
	GetFileFolderInfo /Q/Z strFile
	return modOrCreation ? V_modificationDate : V_creationDate
end

// called after UpdateListboxWaves() to select the listbox items in selList
// whichBox = O: unloaded, 1: loaded
function SelectItems(int whichBox, string selList)
	string strListBox = SelectString(whichBox, "listboxUnloaded", "listboxLoaded")
	
	DFREF dfr = root:Packages:procLoader
	if (whichBox == 1)
		wave selWave = dfr:w_loadedSelection
		wave /T listWave = dfr:w_loadedProcs
	else
		wave selWave = dfr:w_unloadedSelection
		wave /T listWave = dfr:w_unloadedProcs
	endif
	
	int i, numItems
	numItems = ItemsInList(selList)
	for (i=0;i<numItems;i+=1)
		FindValue /TEXT=StringFromList(i, selList) /Z listWave
		if (V_value > -1)
			selWave[v_value] = 1
		endif
	endfor
	ListBox $strListBox, win=LoaderPanel, row=-1
	UpdateButtons()
end

// *** package preferences ***

structure PackagePrefs
	uint32 version
	uint16 options // bit 0: fast search, 1: mouseover, 2: directives, 3: open IM, 4: WM obselete, 5: drag&drop, 6: ignore hide flag, 7: ignore IgorVersion pragma
	uchar sortUnloaded
	uchar sortLoaded
	STRUCT Rect win // window position and size, 8 bytes
	char reserved[128-4-2-1-1-8]
endstructure

// set prefs structure to default values
function PrefsSetDefaults(STRUCT PackagePrefs &prefs)
	prefs.version      = kPrefsVersion
	prefs.win.left     = 20
	prefs.win.top      = 20
	prefs.win.right    = 20 + 500
	prefs.win.bottom   = 20 + 360
	prefs.options      = 7
	prefs.sortUnloaded = 1
	prefs.sortLoaded   = 1
	int i
	for(i=0;i<(128-16);i+=1)
		prefs.reserved[i] = 0
	endfor
end

function LoadPrefs(STRUCT PackagePrefs &prefs)
	LoadPackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
	if (V_flag!=0 || V_bytesRead==0 || prefs.version!=kPrefsVersion)
		PrefsSetDefaults(prefs)
	endif
end

// save window position in package prefs
function SaveWindowPosition(string strWin)
	STRUCT PackagePrefs prefs
	LoadPrefs(prefs)
	
	// save window position
	GetWindow $strWin wsizeRM
	prefs.win.top    = v_top
	prefs.win.left   = v_left
	prefs.win.bottom = v_bottom
	prefs.win.right  = v_right
	
	ControlInfo /W=LoaderPanel popupUnloadedSort
	prefs.sortUnloaded = v_value
	ControlInfo /W=LoaderPanel popupLoadedSort
	prefs.sortLoaded = v_value
	
	SavePackagePreferences ksPackageName, ksPrefsFileName, 0, prefs
end

// ProcedureVersion("") doesn't work for independent modules!
// extract procedure version from this file
function GetThisVersion()
	variable procVersion
	variable noversion = 0	
	Grep /Q/E="(?i)^#pragma[\s]*version[\s]*=" /LIST/Z FunctionPath("")
	if (v_flag != 0)
		return noVersion
	endif
	s_value = LowerStr(TrimString(s_value, 1))
	sscanf s_value, "#pragma version = %f", procVersion
	return (V_flag!=1 || procVersion<=0) ? noVersion : procVersion
endÍ

// *** Button proc pictures ***

// PNG: width= 90, height= 28
Picture pb
	ASCII85Begin
	M,6r;%14!\!!!!.8Ou6I!!!"&!!!!=#R18/!9'oUErZ1?&TgHDFAm*iFE_/6AH5;7DfQssEc39jTBQ
	=U#OG#M5u_NK`L;'g'_`?_5JPn)_d:C()V]%S0h9%3pu4-+e;tk,")Zj,2$t(u:^^!26PP,C#]8!<9
	mWaQ8X-SSBWG/hC(;G*#S&af\J(Vo(h3i=JdEa20h9%Bm_&NRbPIltcZl$<dZ3HKn$CpGHi)smgJt<
	2g*[<'Xa;<t!BdsaPsOO"h=u=#@$P%=i-3?c*Ydi'%Ab)j+3>HX.,7PocZV:W6E^l,aImEkaM]:*'m
	bL`L6pfSJJc63O+3ANeuTl'kA,hOWe93S-G[q*-P6FrcH'TnrT4.>cLlR_KAXsdKjN",$]&0nO+B&R
	">RL]CN;`9._&d][6h'7(^>`1g1:YEMd]WVWriKZkEi[R2%1^O9gfP#?=&qC%a9eZ(_EhMq:"*r[P(
	KYl.*a,BQsH@RWK&3?uap)g+'eI@0A8Pps!/h+sK/DUP6@0D;B_m<dJ-;i7BXaD/V`L"ShK%.6s9po
	k[6<D&EoQbc$+'I\kk\>,>;ai(")=K[B&aRSE%6V,0L-n&sp<gjpr?RaQf%,in^6N4(uiYXT#H/=d<
	="ari;0Dbn$'\G6r;P6kA]QME!=TX0"2;/TMK\=g09eO;AaD^g4$Rg.Y$Or)TjQC*,J.Z>J1TU05@M
	f319f>^S[jh"m35c&+[WA_9$^@M3L;Cr>M0ud?+mCG\D)[i!6,-,m$C_6lSZh0D=ZLU`^-jE(L_/a+
	@f3_e&j>Z+=DRP`\rK#QD&4D5Whg)JX6sW6K#Te`(hHA'h=\a7nTbMA)&$++2'5n%3),AA>ckQTTDn
	[.T'-Fq5Ptp$CkW#c*j.arR/OEDrcr-lYIG0j,k7=iA&VI'^(DJ2.>=jc*pe%TXfX21WitUGK?%te4
	%&aEV)R%LR&uYtK$hc9'lJ(K_DRcVlV"d84.NhbMo_HN78f(sE/-nMMo!p_PPO:X(_.>Mk+uG"-8.R
	WPVW-g@/EC]CSaC>1;k6[M;iS52`f"sQc\`]A>7m@%\'1=+PaWqkMMU>3+#\%bW-',UhH[38oH,Q="
	_UMT\pp3]\I#$;cf3KSY5,^QcpuT(,I#S_&),;?$A@f7XJEo4d:R20_fV17S<CsP'_WLbc,36(1<TC
	k;?Gpi(D7*l>)m1(cR/r&,M(gf0e.1\fs>\Nr;6cL'X^#Nq23XINefNoU6ZRT;&.r>INp.<*Bdb3LR
	g=leH?k^JTssf?)=[jEsr;H]%PBL@fE8TbM)"'uYHIpj>ZC*B%+!p\T"nD"rBLe6_`D=nNZ3^)<7#%
	;#1$Kt@qJT.\7Kn$cWe0fM1',%mr013<Fogs5p1]k;5\:80CqI:\0FYp0=H7k5//)cJLPgpjK,ggo"
	jY1;'bQ8o#EDPObMCT^UF&Ja]lCeZcfkm_3GW+P^2\!rjjEP'+B6tgBGTPFn2<9UoLi"+*5h!Qkb>P
	s<*&L@,dRGb.&RD1V0_s+rJTeuh=B!thc/qRN!\f=HMIX_?uoS,-tn\t)"ZY"XcJX:ZkNH!U-N;/4k
	L:]G4-HpC!W2FJsCN.uBbmek'%s-+MBKcj&k@TCW\ZYfWHsN.5f/V]i`8B"%h.C[:"&KWihV1-(:#U
	#)`o5:L8oMIW*WCSN0fG<^$g2NTai!4^FLeiZ\V5qX/hh$nO1NeAqUNmr"Hno\W2m$h+$1CQDX2=[X
	$8jTD(Tb41&[\doMohnS)K*Gp)kP1Ir$1p2Z0:>:k`A30o?;/LU5sih&JJ'N.Uu?0%ZGS*ol%D10,u
	6bi)KSLqs""Va2GDEX'Qf[eEV$@YhsbOr;Goh%a1n^;RP*;-g?-3`jRUs&6h:l+('.88kD/L\/*K<@
	HiI?G(#$s*3QKE\^+jFk\g_mf"%tr'Z;(TgKJXRs1.'D>SMI->,)ZdWe3"W-"*ToAiq?lB+KniR<3f
	n0KAW?=$BP`As/"\EpZNTu>krVp8.gZCFe0PBm,7ou\bt\fp2G0E'`I^jJBuEkU*WfVrXG[/auCY9M
	heaJ?pV\9N9LKc:C[GiUhe)<lC?,H/A&LJt+O\5[m2Of<&r=S;h#k%mCQQo8r#8G!&^_JZoq9qa)rK
	2(h)J3/Ws_a[I(?cJ5'7laq)RH,VHmVitd)k-R"l8@kmXt_3<bI5[aRp:mDp1`+:ZF<+E4c,u3A[6;
	c+[]1_Z$Q),m+kUYi<+G;Fo)YGOV*OH:LhN+kBZZiI;M`<hY.o]G!QUZ7hP2ak3+0_n9AE11OCH].P
	%J^?U\I''4EL0M2>Q9k2oV=C"n]T<p<VFY'IY"=JZ]\?4R^tOt$>#)[m`ARae]<8r@fMNeg#VN:#3:
	+.li7!9QK=<%k8*_@me^MDp9b#PR*o&l6/!M6iJFE&b`X,-b>ga,15K^H#Dl='`Ri]r0YI/RPb\ieO
	Bd-9[9M4kgk=MrX!dSlOV(fk.LoZ/%(uAE,'!8n:Ts$&Qc'SN]VM9E`=]MgE@M9!@+MP)f7SP>=igc
	r<olZR@9f)mh7V0mBWc7)bQ[Y3]ASqdKZ(=6o5_QS2\h0farNU)f1d*h4>TeCWS#.b\A:3I5Xr$j3E
	:f7@8?k%T]s^!E))hE_N)5lR9V<)jpY)]LYJ6&'s>q`j;pX7,UPN#_lu!!#SZ:.26O@"J
	ASCII85End
end

// PNG: width= 90, height= 30
Picture cog
	ASCII85Begin
	M,6r;%14!\!!!!.8Ou6I!!!"&!!!!?#R18/!3BT8GQ7^D&TgHDFAm*iFE_/6AH5;7DfQssEc39jTBQ
	=U%adj95u_NKeXCo&'5*oW5;QerS2ddE(FP;'6Ql#oBiuS1dW/'TBu9uWoJd,<?@,Ku[OT_*C6>N&V
	5fgKO-Z!0m]BHZFE$kSY/b+).^UIV80KE-KEW"FQ9_,eF^bT7+UPEE3;1R;IlL+VJZGMtc/)57aYSB
	X^5kLMq7Sp9q0YBG@U214ph]9b)&eA4@(!#Ip%g0/F0_UH>\E'7M.=ui]mI6`m'G"T)B'OA[IVmQDm
	=?Eat`-Qd[5bR1c#F]`cT(Se*XgJen[DI=<Y8YF7N!d-X+iY2pq;SVDt;siWS_bs$!E]WT^pRhs]aH
	M%csq\akmJ4tMo<HGU1WqT,5cib%XrHM`>*]/&U:](RQ7QDj\u&#cMNi8>`?8;-?rCcXX>+1^gW10F
	n!h-2G-M;R\a+TD2op'?`R]"%W!DpJmO],i(@-(/F?,;L7Vkt%YOa,f\8+CV@lHVOEMgMZPnh$6>!V
	MTYBm^8f[O,?Z$2MnH6.T'JWSM4HtSissRo-):4d:ecoe5Tn^(gUEQm+o94@;L(/[91%aXk:!pP;mm
	\kh$s.7qbe%=-p1eBtDs*CHp:+$CUY\0A,jM0:4J2&pY-HWBG?nb`"BE/M-#*)+E?I*C/(r;J]APNh
	3!Ea<(u)`o?0R`ma=QV<n?GV/s3:I0Wf2_M0p@:__T%OEl+sL@10K8&ViQgR(0Q3qMLYA':/iba:,;
	]Y$@ACMV&9b[fD4A`Vq5+A!37VD0na`;0#fWNWKq#f5N>Mt)$S['[2:?=(p2$Q$$NX_cXoJ`iVOcHm
	Rb+"_b#*b4@tp)Xq9r*1_<^IVlpMJ=kE>MhiHa2]]q9<d*4(lA_8$4ej2NM5Z!#`oc=+Ttk-]%D5"O
	Yiu,o$V/I<=@2fN3Ds,PNfIEnqn6C?^[OYDs4q2k*s6TFu+@1>SKUmdko@B5>Pp)-]8`l_Ig,/1c.T
	K'Z+asa)qDc*mqZAKmijlOd;;&H$MEMWY1:\q<G#aaNVlho?TWCGL35!G658MH$RpQ,/[:S#==eP-@
	T+s%'h-7&.0)\eW6j@1gNW'FYlgRIid1g1dP0.MtL)"o@*4A+B&XU\9bRSJWg?B!keI%b6T7FS'?W(
	@7j-a-n[,Adkh%U((6"oN9G"iRV$rmb0"2lqXk08!N&V_b13Oo*tc)h=[L]E@W<ihr:]%Dbs-*cCbc
	T^`<b81D1(d_gue7JX)rMl-Q!ag!0.a4mbCL5'MQH2X<p3`1nA#69QNiWDnN[J^:kIm`JPCXo#W8lo
	?KFN_dOMp#7s\i!;q:1Vb`q^Za5j'0Sg8ALVn\tm:OM*.G/IF;=K4S+O/0U]^`u\O,i'69Y*0'f,SH
	:Kp)mH3.EQ3hDf%?f;W[LbJbR7R5Lb$8U4I7,[8ZM7fU7H5(>6BlSlr1cG6"263q!"YH.\`>aLYN(!
	e3hZYm666$Mq_c(_GHO?%CE(rnI-UV=I6M\e$%CXt$`9q"IB8d`/4e)0&Dcf`43oobf6MqdV?d>\73
	X/dI(2jZ+#[Nt'sQg5KRFl_^rEVX>cZ5h3g1gR#*IhTZf+KqmsVrW[`UcE6>MM*N0fMcT[S!:h'=ju
	EaL@5>OfU]WXd+d/JISLYCQZ_BPkB$IiAYUBq1l^ecC4a8EYJ'WJ,pak5V59k6$F23m\(d</D&W$.c
	&:n64N(\`j0%h;m7iB=j)(R^kh8BUCfn<.JP[26K1F7\>0JOc56jWCOX(6k=i$m^+A%<G5QZ.i$e8]
	01cU/k$R?&@c[1P\L>tK[OkSMmQ7lT_puI&4&#-'R99q+p;3\Sn`Ic2G%kDj/"N6oB<E0?Z6Kl"@,Y
	?4P5G,Nu\q^LTs(Vj(g3YCoZEl;W9!T)>ePG,S!10n_Y\rP(FBr90.J3;jVP^MYp<-ML6>(=)*bFsB
	&uWX-VWg'<P8G@$+\A?J1@G'F,/Z.<q/;.,=8I?;gFp>>;IjEQPE_:7`.R+3bElA@DB6<k@ksJ9lg(
	=CVM=g<G(^E#SiiFHZ8.qF-^ppkE&\[U*_);<'Lfk*Fq]^#W339=Q'IWpj"KgUWU]1,ETW<&G^OtOH
	MkGSX1r0@A!=Gd&>m2;$s(nF8b!K?8R`eZjO8V38_I$<1AcnP!)B*`T#3.X=hE[s8PMoFf*/H2@32u
	N'Len`GUk?n:K*@=9gMMi1ES9gF7flQcCD`2n^,h:`S5=GNQ^G#@^/a:?]W`PV50n4Xue>QVk8E1=]
	lWKB?uV(SiXjL_hVC,G-+W-bHd)[C`^u(BPM:VV58ltJcZ8d$CEhp-Ek)Qb.'^'f)4eT`-]7*:O[1>
	WO?>HRRZ3%&Aum4gNS.jU=.^S*B#H\'0H3Z)tG>O;)WW1_IO09[>@PI4Y3R7J0]^!UgQhj+u1-,O_#
	4q7aj2?Y5.V]ph=D*J`g8?n%JH:8P)K%MLrfVTs(X1]A:d+mFtdNBG"";'8siHNZC4&bKJq`%mNb7r
	SW;=`2-+n=L)HDOsFHoS$CX_6m<3W758iRSt7"9?7u`s%5^"&NsfV]&fKiR),nod"F>!-ZZj257RH'
	BCpkU2>pgE:BYW?=m,Fa:(R[)N0g+8U_d1?h6?87.J?1.S8QM+N.?c/5H^[K9Qq/KSe*09PFi*)kOs
	Cpa8151hB![K\`b9:/2t#`0h)TQGGW^_mOCaj@jCA@uU*q95,uIW@7!X&<O\"P/_=YGg"!\PP(fcMs
	XX8]B01I/5GVm/oHal.qLkA^e?r`*cq5<6cMYpEN]i/%/Vl(p'^fIMd8C\oHLnT1!)7oQm]mKeID*1
	Nn<=:2#ZkEe#YpHHH,[1j)T%7I4;@/)4cu_Q1)PbZM:[@i"UEI%;r>p03XoVZ_RkU=+$'7#=USG/bL
	8-+m<=>h,iqN=A8kNQ3E0-ce+TlUO7L$\:&5CW07\^Y5(=Lpj3bn5fXf]+hD?H]7Wq?&[-c"7hNK0#
	/)B'Mj<HT8ma3CurC,*bh\'aHR7Z[QbX-ZmAk800m):lpO:?P+(!;TY'R\j"AmjWGZ^'4n^bf?W3J@
	>&NBK1HqWkVhSJ@2:%U$9,hq:d,G1cCmI-TcrP\J*Ut[2@6?BcK3XN6]^DH?sm>]m;PWk0+t]M3*pb
	_i5ToaNr1&dko4ib1O7G-^#a3R58IWd+6c;6ULrU<E06&]A7B"H::^+p=jM"Cht@E-\k9W-F%RN7[f
	g9`s&hll-^kfa*7KZfMd!Yc("^(?tb@(G_h8BF>IEA!<RhTlND,!db&r!Y3m>2$M&6dHhJmnR;"(TN
	7k9deNFM:rt_%M:_\bI5MO2h<A0N#R:ZSrC"&ps\iY*%&:=-;@IrX+"G9!l_&sOI?=_'7)$hsk)[Og
	CfZ9V$@mNB]AS#G_>V6^Z_/)$iF?19\*_+U8'Lh!@O$@74\ogtP<K2K,l#,B;0e6NJ*#oS35C1Gqa?
	[/#EMYbdp\'g083t^Hm.OC']=?TCh[+\@'/CD^$lb9k>s0TnKIaqh4fI!&lDq*\4*U*,*??/2AnId;
	.P@%q^Y_gV7L#<Y@CP!Nm,EKn=&i8<r@!PTa5]H_PQ](fBqn;iMUE,Pl5E-<#!*W9G0Gifi6\]*<o4
	FnfrX,WF^_\F$1nF]^fU-bLO!=bm%5lG.k3$IWMqUu'c@l,R*B4I#7$5RGsBAo;S"q!W&r%8C2/"PK
	bsa<4<J73edT0;DW#W4)GM@uFDONL"9R+_N^([T!(k&'aJc*VLHV&^R;!(_#4Z$g7@#[rnEd4b][m6
	"s2Cf6FX#Yth+![-Bc9;DCc35!#ZOe]>R5l%A3s9r*"E3cZ^HAq!@!X3ZLR*.A7oQ8om.\p[kV)RYJ
	YS_VJ,ke(!91A):`eg`H3<A0s%C//=2RBq,pC2:S]*\P@U_Oa*3.T[g!Hf.uI$^Z574:Ig+a&Rh&b)
	g:i!IBPVCY]Y$?Mc-nM/==coe'#A=jP*M;$C2,4.LP+[KAA[:ZiG]VW6ipmf;5gRtUogbYmG#,M=Y+
	23\;kLm>:;.qMlKqn)uClJ![.i',6Vo?VRu"<5C39QHia.rIUXNcq/492/E<t4:q=5j]#0#BT^2C8R
	r=7NOaA&EGCimE'I"(o(b72sE7cRPmWBnj]tHh/;(=(Hs(iTH$*GN/REC6P4"-OQ)1Z*S9OlNXqYG,
	/qFo"%Xt5WcH)DMDo_@'gn2mp\l)\(b!Y0Pa!2n,Nj%-R@Yjn?WT$E#t(FUbj=.R08ON,:0qYL%:/M
	0]<Q1)2)G':0@s*h8ZZ<4MLPu4'A3d&SI6l9j+cCI%0ltDh?G(&13<0N\Q=?t??;p9Q8$59_n3Rli=
	5b`N"1`$#>;[JNr+$!*^fli%qGlHBaGd#r_gndbH0<a<pR0sE5L0:j)I_t)\EH/7Wqt8QJMd<r<&WK
	8J3cuoH9hij#22_bS-?/1q+bUC@(DjDc_1Dg*LCYK([C$_m"OB=44C54XF6CiRHM)#JSik-Qi#lgdX
	C:AAV;n0(-%L_0m!T,"d5MVG@HhTKZ87K.p%WH^Xh3kCpXcTY<@p]%p!H!PcGm8Ma`dWI-i:%O`.@A
	\EMhGlp>Rk7O<G2m`*r,h[u\8;4r,bUaQp%EDN*J]D4B1hFXuppq^tpModARV5%<QlNN?L%hAEkIlW
	/#`^]Bs#-d.f-97RH2#l<o@ZXZ&T?iRH-<%N9SU+'so7AAgW2mil0fWaM)A,6@)4Rp@WFC0@Y,uIN:
	5uCJkMPAJFd6VVd/UR6[rI=?0fVgq/kI,s^(YARDN54WD\PD#"bV<C5XKS<`]Fql#m@)ul]O#Nn]-2
	[lB);<HM,sPARkJs:;d4_S!/nh?qGe7@I:AsnMi7DjM_D$2XW>fsY^ZQI8$;`nm!f"mg^^mq'BNs/!
	!!!j78?7R6=>B
	ASCII85End
end
